# -*- coding: utf-8 -*-
"""
Synchronization tests, using py.test. Run as:

$ easy_install -U py
$ py.test
"""
from threading import Thread, Lock, RLock
from time import time, sleep

from langacore.kit.concurrency import synchronized

SLEEP_AMOUNT=0.1 #seconds

def test_simple_synchronization(sleep_amount=SLEEP_AMOUNT):
    @synchronized
    def simple_synchro(result):
        for i in xrange(10):
            result.append(i)
            sleep(sleep_amount)
    
    class SimpleThread(Thread):
        def __init__(self, result):
            Thread.__init__(self)
            self.result = result
        def run(self):
            simple_synchro(self.result)

    result = []
    threads = [SimpleThread(result) for i in range(10)]
    for t in threads:
        t.start()
    for t in threads:
        t.join()
    assert result == range(10) * 10


def test_shared_synchronization(sleep_amount=SLEEP_AMOUNT):
    lock = Lock()
    
    class SharedLockThread(Thread):
        def __init__(self, result):
            Thread.__init__(self)
            self.result = result

        @synchronized(lock=lock) 
        def run(self):
            for i in xrange(10):
                self.result.append(i)
                sleep(sleep_amount)

    result = []
    threads = [SharedLockThread(result) for i in range(10)]
    for t in threads:
        t.start()
    for t in threads:
        t.join()
    assert result == range(10) * 10


def test_reentrant_synchronization(sleep_amount=SLEEP_AMOUNT):
    lock = RLock()

    class ReentrantThread(Thread):
        def __init__(self, result):
            Thread.__init__(self)
            self.result = result

        @synchronized(lock=lock) 
        def run(self, i=0):
            if i > 9:
                return
            
            self.result.append(i)
            sleep(sleep_amount)
            self.run(i + 1)

    result = []
    threads = [ReentrantThread(result) for i in range(10)]
    for t in threads:
        t.start()
    for t in threads:
        t.join()
    assert result == range(10) * 10


def test_filebased_synchronization(sleep_amount=SLEEP_AMOUNT):
    class FilesystemLockThread(Thread):
        def __init__(self, result):
            Thread.__init__(self)
            self.result = result

        @synchronized(path='./test.lock') 
        def run(self):
            for i in xrange(10):
                self.result.append(i)
                sleep(sleep_amount)

    result = []
    threads = [FilesystemLockThread(result) for i in range(10)]
    for t in threads:
        t.start()
    for t in threads:
        t.join()
    assert result == range(10) * 10



# -*- coding: utf-8 -*-
"""
A reusable memoization decorator implementation. It is using a cache of pickled arguments
"""
 
import cPickle as pickle
import sys
from time import time
from functools import wraps
 

DEFAULT_MEMOIZE_MAX_SIZE=256

 
def memoize(func=None, update_interval=300, max_size=DEFAULT_MEMOIZE_MAX_SIZE, skip_first=False):
    """ memoization decorator. 
    
        update_interval - time in seconds after which the actual function will be called again 
        max_size - maximum buffer count for distinct memoize hashes for the function 
        skip_first - if True, the first argument to the actual function won't be added to the memoize hash 
    """

    # the decarator can be used with an argument as well as without any 
    if func is None:
        def wrapper(f):
            return memoize(f, update_interval=update_interval, max_size=max_size, skip_first=skip_first)
        return wrapper
 
    cached_values = {}
    lru_list = []
 
    @wraps(func)
    def wrapper(*args, **kwargs):
        if skip_first:
            key = pickle.dumps((args[1:], kwargs))
        else:
            key = pickle.dumps((args, kwargs))
 
        if key in cached_values:
            # move argument to the end of the LRU list
            lru_list.append(lru_list.pop(lru_list.index(key)))
            
            # get the buffered values and check whether they are up-to-date
            result, acquisition_time = cached_values[key]
            if update_interval and time() - acquisition_time > update_interval:
                del cached_values[key]

        if key not in cached_values:
            cached_values[key] = (func(*args, **kwargs), time())
            lru_list.append(key)   
 
            # clear the least recently used value if the maximum size 
            # of the buffer is exceeded
            if max_size is not None and len(lru_list) > max_size:
                del cached_values[lru_list.pop(0)]

        return cached_values[key][0]
 
    return wrapper



# -*- coding: utf-8 -*-
"""
A reusable synchronization decorator implementation. It is using 
threading locks or filesystem-based locks to synchronize
subsequent calls of the specified functions. The former kind of
lock is reentrant, the latter is not.

For filesystem-based locks the module is using Skip Montanaro's
``lockfile`` library, compatible with Windows and POSIX environments.
"""
from __future__ import with_statement

import sys
from threading import RLock
from lockfile import FileLock
from functools import wraps
 

def synchronized(func=None, lock=None, path=None):
    """ synchronization decorator. 
    
        lock - the user can specify a concrete lock object to be used 
               with this specific synchronization decorator. This is 
               useful when a group of functions should be synchronized
               together.
        path - instead of using threading-based locking, file-based
               locks may be used instead. Beware, these are radically
               less performant than threading locks.
    """

    # the decarator can be used with an argument as well as without any 
    if func is None:
        def wrapper(f):
            return synchronized(f, lock=lock)
        return wrapper
 
    # threading lock
    if path is None:
        _lock = lock if lock else RLock()

        @wraps(func)
        def wrapper(*args, **kwargs):
            with _lock:
                result = func(*args, **kwargs)
            return result
    # file-based lock
    else:
        @wraps(func)
        def wrapper(*args, **kwargs):
            _lock = FileLock(path)
            while not _lock.is_locked():
                _lock.acquire()
            result = func(*args, **kwargs)
            _lock.release()
            return result
    
    return wrapper



# -*- coding: utf-8 -*-
"""
Memoization tests, using py.test. Run as:

$ easy_install -U py
$ py.test
"""
from time import time, sleep

from langacore.kit.cache import memoize


def test_memoization_update_interval():
    @memoize(update_interval=4)
    def current_time():
        return time()

    t = current_time()
    sleep(1)
    assert t == current_time()
    sleep(1)
    assert t == current_time()
    sleep(3)
    assert t != current_time()


def test_memoization_max_size():
    @memoize(max_size=2)
    def current_time(arg):
        return time()

    t1 = current_time(1)
    sleep(1)
    t2 = current_time(2)

    sleep(1)
    assert t1 == current_time(1)
    assert t2 == current_time(2)

    t3 = current_time(3)
    sleep(1)
    t4 = current_time(4)

    sleep(1)
    assert t3 == current_time(3)
    assert t4 == current_time(4)
    assert t1 != current_time(1)
    assert t2 != current_time(2)
    
    sleep(1)
    assert t3 != current_time(3)
    assert t4 != current_time(4)



# -*- encoding: utf-8 -*-

from setuptools import setup, find_packages

setup (
    name = 'langacore.kit.common',
    version = '0.1.4',
    author = 'LangaCore, Lukasz Langa',
    author_email = 'support@langacore.org, lukasz@langa.pl',
    description = "A library of various simple common routines that keep being " \
                  "rewritten all over again in every project we're working on.",
    long_description = '',
    keywords = '',
    platforms = ['any'],
    license = 'GPL v3', 
    packages = find_packages('src'),
    include_package_data = True,
    package_dir = {'':'src'},
    namespace_packages = ['langacore', 'langacore.kit'],
    zip_safe = True,
    install_requires = [
        'setuptools',
        'lockfile',
        ],
    
    classifiers = [
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License (GPL)',
        'Natural Language :: English',
        'Operating System :: POSIX',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows :: Windows NT/2000',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.5',
        'Programming Language :: Python :: 2.6',
        'Topic :: Software Development :: Libraries :: Python Modules',
        ]
    )
